#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#!/opt/rb/var/www/rb-rails/script/rails runner
#
# Copyright (c) 2024 redBorder Networks
# Authors:
#   Miguel Negrón:         manegron@redborder.com
#   David Vanhoucke:       dvanhoucke@redborder.com
#
###########################################################################################################

require 'yaml'
require 'net/http'
require 'tempfile'
require 'digest/md5'

module Redborder

  class ModelDownloadFailed < StandardError; end;
  class DiskUsageError < StandardError; end;

  class ModelsUpdate

    def self.start(verbose=false, source_id=nil, limit=10)

      @content = ""

      raise DiskUsageError if not_enough_free_space?('/', limit)

      now = Time.now

      sources = source_id.nil? ? ModelSource.where(enable: true) : ModelSource.where(id: source_id)

      @config = YAML.load_file("#{Rails.root}/config/llm_models.yml")

      logit "Node " + `hostname`.chomp!

      logit ""

      logit "Starting download of models.."

      @failed_download = false
      
      # Download models from each source
      sources.each do |source|

        begin       
          new_model      = false    # if true we have new rules and we will create a new RuleDbVersion.
          download_files = false
          @warn_version   = false

          logit ""
          logit "############################################################################"
          logit "       Model source: #{source.clean_name}"
          logit "############################################################################"
          
          raise DiskUsageError if not_enough_free_space?("#{@config["cache_dir"]}/", limit, source.model_size)

          if source.cached_uri.nil? # if no cached_uri is stored, force download files
            download_files = true
          else
            md5_file = "#{@config["cache_dir"]}/#{source.clean_name}.md5"
            File.delete(md5_file) if File.exists?(md5_file)

            if source.md5url.present? 
              logit "#{I18n.t 'lib_redborder.downloading'} #{source.full_md5url}" if verbose
              logit "  * #{I18n.t 'lib_redborder.saving'} #{md5_file}" if verbose

              download_file(source.full_md5url, md5_file)
              next if @failed_download
              # Get only the md5 content, ignoring the name of the model in the file
              md5_file_content = File.open(md5_file) {|f| f.readline}.gsub(/^"/, '').chomp[/\b[a-f0-9]{32}\b/] 
              
              if md5_file_content != source.md5 # if md5sum is different, force download files
                logit "     - Cached tgz MD5 -> #{source.md5}" if verbose
                logit "     - #{source.full_url} MD5 -> #{md5_file_content}" if verbose
                download_files = true
              else
                logit "  * #{I18n.t 'lib_redborder.cached_file'}" if verbose
                logit "     - Cached tgz MD5 -> #{source.md5}" if verbose
                logit "     - #{source.full_url} MD5 -> #{md5_file_content}" if verbose
              end
            else
              download_files = true
            end
          end

          if download_files
            tmp_file = "#{@config["cache_dir"]}/#{source.clean_name}-tmp"

            logit "#{I18n.t 'lib_redborder.downloading'} #{source.full_url}" if verbose
            logit "  * #{I18n.t 'lib_redborder.saving'} #{tmp_file}" if verbose

            download_file(source.full_url, tmp_file)
            next if @failed_download

            #check if the file is correct
            if defined?(md5_file_content) and md5_file_content and md5_file_content != Digest::MD5.hexdigest(File.read(tmp_file))
              logit "  * #{I18n.t 'lib_redborder.not_match'}" if verbose
              @failed_download = true
              next # go to next source
            elsif defined?(md5_file_content) and md5_file_content
              new_model = true
            elsif source.cached_uri.nil? # source has no md5url and no cached_uri
              new_model = true
            elsif source.md5 != Digest::MD5.hexdigest(File.read(tmp_file))
              new_model = true
            end
          end

          next if !new_model
         
          FileUtils.mv(tmp_file, "#{@config["cache_dir"]}/#{source.clean_name}")
        
          source.cached = File.open("#{@config["cache_dir"]}/#{source.clean_name}")
          source.md5 = Digest::MD5.hexdigest(File.read("#{@config["cache_dir"]}/#{source.clean_name}"))
          source.save

          if Setting.llm_selected_model.nil? || Setting.llm_selected_model.empty?
            Setting.llm_selected_model = source.name
            Setting.save
          end

        rescue DiskUsageError => e
          logit "#{I18n.t 'lib_redborder.no_enough_space_disk'} (ERROR: #{e})" if verbose
          return @content
        rescue Exception => e
          logit "#{I18n.t 'lib_redborder.error_source'} #{source.name} (#{e})"
          e.backtrace.each do |b|
            logit b
          end

        ensure
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}")
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}-tmp")
          FileUtils.rm_f("#{@config["cache_dir"]}/#{source.clean_name}.md5")
          FileUtils.rm_rf("#{@config["cache_dir"]}/#{source.clean_name}")
        end
      end

      logit "\n#{I18n.t 'lib_redborder.warning'} #{Sensor::SNORT_VERSIONS.join(",")}\n" if @warn_version

      logit "#{I18n.t 'lib_redborder.exiting'}" if verbose

      raise ModelDownloadFailed if @failed_download
      return @content
    
    rescue ModelDownloadFailed => e
      logit "#{I18n.t 'lib_redborder.error_model_download'} (ERROR: #{e})"
      return @content
    rescue DiskUsageError => e
      logit "#{I18n.t 'lib_redborder.no_enough_space_disk'} (ERROR: #{e})" if verbose
      return @content
    rescue Exception => e
      logit "#{I18n.t 'lib_redborder.error_updating'} (ERROR: #{e})"
      @failed_download = true
      return @content
    end

    def self.get_failed
      @failed_download
    end

    def self.find(dir, filename="*.*", subdirs=true)
      Dir[ subdirs ? File.join(dir.split(/\\/), "**", filename) : File.join(dir.split(/\\/), filename) ]
    end

    def self.logit(msg)
      STDOUT.puts "#{msg}"
      @content += msg + "\n"
    end

    def self.set_proxy_options
      if Setting.proxy? and Setting.proxy_address.present? and !url.start_with?"https://redborder.s3"
        proxy_port = Setting.proxy_port.to_i
        proxy_port = 8080 if proxy_port <= 0
        proxy_addr = Setting.proxy_address
        proxy_addr = ("http://" + proxy_addr) if !proxy_addr.start_with?"http://" and !proxy_addr.start_with?"https://"
        proxy = URI.parse(proxy_addr + ":" + proxy_port.to_s)

        if Setting.proxy_user.present?
          proxy_options = {proxy_http_basic_authentication: [proxy, Setting.proxy_user, Setting.proxy_password]}
        else
          proxy_options = {proxy: proxy}
        end
      else
        proxy_options = {proxy: false}
      end
      proxy_options.merge(ssl_verify_mode: OpenSSL::SSL::VERIFY_NONE, read_timeout: 10)
      return proxy_options
    end

    def self.download_file(url, local_file)
      File.delete(local_file) if File.exists?(local_file)

      url = ("http://" + url) if !url.start_with?"http://" and !url.start_with?"https://" and !url.start_with?"ftp://"

      proxy = set_proxy_options

      begin
        File.open(local_file, "wb") do |writeOut|
          open(url, proxy) do |remoteFile|
            buffer = ""
            while remoteFile.read(4096, buffer)
              writeOut.write(buffer)
            end
          end
        end
      rescue Exception => e
        logit "(ERROR: #{e})"
        @failed_downlod = true
      ensure
        logit "Download attempt finished."
      end
    end

    def self.not_enough_free_space?(path, limit, model_size = 0)
      disk_used_pcent = `df --output=pcent #{path} | tail -1`.strip.delete('%').to_i
      # If model_size is passed, check disk usage after download
      if model_size > 0
        total_disk_size = `df --output=size -B1 #{path} | tail -1`.strip.to_i
        avail_limited_size = total_disk_size - total_disk_size * ((disk_used_pcent + limit).to_f/100)

        return true if (100 - disk_used_pcent) < limit || avail_limited_size < model_size
      else
        return true if (100 - disk_used_pcent) < limit
      end
      
      return false
    end

  end
end
